/* TransConnect : A function imposter to allow transparent connection to the
 * internet over HTTPS proxy like squid.
 * 
 * Copyright (C) 2001, Dwivedi Ajay kumar <a_dwivedi at biosys.net>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * or visit http://www.gnu.org/copyleft/gpl.html
 * 
 * Credits:
 *    Raja Subramanian : for pointing me to the imposter method.
 *    
 *    Linux India Mailing List : A place where all your questions get 
 *         answered within no time.
 *    
 *    Jared D. McNeil  : for helping me with port to NetBSD
 *    
 *    Harald Fielker and Sebastian Stark : for providing patches for
 *    					                   OpenBSD.
 *         
 *         
 */

/* This set this to 1 if you want to TCONN_DEBUG your connection */

#ifdef TCONN_DEBUG
#undef TCONN_DEBUG
#endif /* TCONN_DEBUG */
#define TCONN_DEBUG 0

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<strings.h>
#include<sys/types.h>
#include<sys/socket.h>
#include<netinet/in.h>
#include<arpa/inet.h>
#include<netdb.h>
#include<unistd.h>
#include<errno.h>
#include<dlfcn.h>
#include<pwd.h>
#include<signal.h>
#include<fcntl.h>


 /* Maximum no of localnetworks you can have in config file */
#define MAX_LOCAL_NETS 10

 /* Friendly states */
#ifdef TCONN_FALSE
#undef TCONN_FALSE
#endif /* TCONN_FALSE */
#define TCONN_FALSE 0

#ifdef TCONN_TRUE
#undef TCONN_TRUE
#endif /* TCONN_TRUE */
#define TCONN_TRUE 1

/* Some people using RH7 and Suse7.1 are having problems without this */

#ifndef RTLD_NEXT
#define RTLD_NEXT       ((void *) -1l)
#endif /* !RTLD_NEXT */

/* Comptibility with BSD */
#ifdef _BSD_HACK_
#include "libcpath.h"
#endif /* _BSD_HACK_ */

 /* This TCONN_MATCH is taken directly from glibc resolver. */

#ifdef TCONN_MATCH
#undef TCONN_MATCH
#endif /* TCONN_MATCH */
#define TCONN_MATCH(line, name) \
      (!strncmp(line, name, sizeof(name) - 1) && \
       (line[sizeof(name) - 1] == ' ' || \
       line[sizeof(name) - 1] == '\t'))

struct my_network
{
  int valid;
  struct in_addr subnetwork;
  struct in_addr subnetmask;
};

/* Just to make sure this never conflicts with any function call */
void ajayd_tconn_base64_encode (const char *s, char * p);

int
connect (int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen)
{

  char proxyserv[20];		/* The proxyserver to use. */
  int proxyport;		/* The proxyport on proxy server */
  char proxyauth[80];		/* Some proxies like mine need proxy authorisation. */
  char proxyuser[10];
  char proxypass[20];
  struct my_network localnet[MAX_LOCAL_NETS];
  char buf[100], *cp;
  FILE *fp;			/* For opening config file */
  static int (*func) ();	/* Ah.. The function we are overriding :-) */
  int lnnum = 0;		/* A number to count localnetworks. */
  int proxyservset = TCONN_FALSE;
  int proxyportset = TCONN_FALSE;
  int proxyuserset = TCONN_FALSE;
  int proxypassset = TCONN_FALSE;
  int localnetset = TCONN_FALSE;
  int isonlocalnet = TCONN_FALSE;
  int optval;
  int optlen;
  int uid;
  char configfile[100];
  struct passwd *pwent;
#ifdef _BSD_HACK_
  static void *handle;
#endif /* _BSD_HACK_ */

  /* This is the core. */
  /* Set func to the next instance of connect */

  if (!func){
#ifdef _BSD_HACK_
    handle = dlopen(TCONN_LIBC_PATH, RTLD_LAZY);
    func = (int (*)()) dlsym (handle, "connect");
#else
    func = (int (*)()) dlsym (RTLD_NEXT, "connect");
#endif
    }

  optlen = sizeof (int);
  uid = getuid ();

  if ((pwent = getpwuid (uid)) != NULL)
    {
      strncpy (configfile, pwent->pw_dir, 80);
      strncat (configfile, "/.tconn/tconn.conf", 19);
      if (TCONN_DEBUG) fprintf (stderr, "Configfile is %s\n", configfile);
    }
  else
    {
      /* Ok Ok.. Goto's are bad, But is just a bit easy here. May change Later. */
      fprintf (stderr, "You do not exist\n");
      goto directconnect;
    }

  if ((fp = fopen (configfile, "r")) != NULL)
    {
      if (TCONN_DEBUG) fprintf (stderr, "reading %s\n", configfile);
    }
  else
    {
      goto directconnect;
    }


  getsockopt (sockfd, SOL_SOCKET, SO_TYPE, &optval, &optlen);

  /* We want only Internet TCP sockets to use our method. If you remove this,
   * all your file sockets and X appls will stop working (At least in this shell). */

  if (optval != SOCK_STREAM || serv_addr->sa_family != AF_INET)
    {
      if (TCONN_DEBUG) fprintf (stderr, "Wrong socket type for routing\n"
	       "optval == %d sa_family == %d optlen == %d \n"
	       "whereas \n"
	       "stream == %d AF_INET == %d optlen ==%d \n ",
	       optval, serv_addr->sa_family, optlen,
	       SOCK_STREAM, AF_INET, sizeof (optval));

      goto directconnect;
    }

  while (fgets (buf, 100, fp) != NULL)
    {
      if (*buf == '#')
	continue;

      if (TCONN_MATCH (buf, "proxyserv"))
	{
	  cp = buf + sizeof ("proxyserv") - 1;
	  while (*cp == ' ' || *cp == '\t')
	    cp++;
	  if ((*cp == '\0') || (*cp == '\n'))
	    continue;
	  if (index (cp, '\n') != NULL)
	    *index (cp, '\n') = '\0';

	  if (inet_addr (cp) != -1)	/* We are safe here as proxy can't be 255.255.255.255 */
	    {
	      struct in_addr proxy_addr;
	      proxyservset = TCONN_TRUE;
	      proxy_addr.s_addr = inet_addr (cp);
	      strncpy (proxyserv, inet_ntoa (proxy_addr), 16);
	      if (TCONN_DEBUG) fprintf (stderr, "proxyserver is %s\n", proxyserv);
	    }
	  else
	    fprintf (stderr, "%s: proxyserv: "
		     "Only IP addresses are allowed\n", configfile);
	}

      if (TCONN_MATCH (buf, "proxyport"))
	{
	  cp = buf + sizeof ("proxyport") - 1;
	  while (*cp == ' ' || *cp == '\t')
	    cp++;
	  if ((*cp == '\0') || (*cp == '\n'))
	    continue;
	  if (index (cp, '\n') != NULL)
	    *index (cp, '\n') = '\0';

	  if ((proxyport = atoi (cp)) != 0)
	    {
	      proxyportset = TCONN_TRUE;
	      if (TCONN_DEBUG) fprintf (stderr, "proxyport is %d\n", proxyport);
	    }
	}

      if (TCONN_MATCH (buf, "proxyuser"))
	{
	  cp = buf + sizeof ("proxyuser") - 1;
	  while (*cp == ' ' || *cp == '\t')
	    cp++;
	  if ((*cp == '\0') || (*cp == '\n'))
	    continue;
	  if (index (cp, '\n') != NULL)
	    *index (cp, '\n') = '\0';
	  /* The proxyuser is in quotes */
	  if (*cp != '\"')
	    {
	      fprintf (stderr,
		       "%s : proxyuser does not start with double quote \n",
		       configfile);
	      continue;
	    }
	  cp++;
	  if (rindex (cp, '\"') != NULL)
	    *rindex (cp, '\"') = '\0';
	  else
	    {
	      fprintf (stderr,
		       "%s : proxyuser does not end with double quote \n",
		       configfile); continue;
	    }
	  proxyuserset = 1;
	  strncpy (proxyuser, cp, 9);
	  if (TCONN_DEBUG) fprintf (stderr, "proxyuser is \"%s\"\n", proxyuser);
	}

      if (TCONN_MATCH (buf, "proxypass"))
	{
	  cp = buf + sizeof ("proxypass") - 1;

	  while (*cp == ' ' || *cp == '\t')
	    cp++;
	  if ((*cp == '\0') || (*cp == '\n'))
	    continue;
	  if (index (cp, '\n') != NULL)
	    *index (cp, '\n') = '\0';
	  /* The proxypass is in quotes */
	  if (*cp != '\"')
	    {
	      fprintf (stderr,
		       "%s : proxypass does not start with double quote \n",
		       configfile); continue;
	    }
	  cp++;
	  if (rindex (cp, '\"') != NULL)
	    *rindex (cp, '\"') = '\0';
	  else
	    {
	      fprintf (stderr,
		       "%s : proxypass does not end with double quote",
		       configfile);
	      continue;
	    }
	  proxypassset = 1;
	  strncpy (proxypass, cp, 19);
	  if (TCONN_DEBUG) fprintf (stderr, "proxypass is \"%s\"\n", proxypass);
	}
      if (TCONN_MATCH (buf, "localnet") && lnnum < MAX_LOCAL_NETS)
	{
	  cp = buf + sizeof ("localnet") - 1;
	  while (*cp == ' ' || *cp == '\t')
	    cp++;
	  if ((*cp == '\0') || (*cp == '\n'))
	    continue;
	  if (index (cp, '\n') != NULL)
	    *index (cp, '\n') = '\0';
	  if (index (cp, '/') != NULL)
	    {
	      localnetset = TCONN_TRUE;
		/* If error inet_ntoa return is same as for  255.255.255.255 */
		{
		  localnet[lnnum].subnetmask.s_addr =
		    inet_addr (index (cp, '/') + 1);
		  if (TCONN_DEBUG) fprintf (stderr, "Subnetmask is %s\n",
			   inet_ntoa (localnet[lnnum].subnetmask));
		}

	      *index (cp, '/') = '\0';
		{
		  localnet[lnnum].subnetwork.s_addr = inet_addr (cp);
		  if (TCONN_DEBUG) fprintf (stderr, "Subnetwork is %s\n",
			   inet_ntoa (localnet[lnnum].subnetwork));
		}
	    }


	  localnet[lnnum].valid = TCONN_TRUE;
	  localnet[lnnum + 1].valid = TCONN_FALSE;
	  lnnum++;
	}

    }

  for (lnnum = 0; localnet[lnnum].valid == TCONN_TRUE; lnnum++)
    {
      if (
	  (((struct
	     sockaddr_in *)
	    serv_addr)->sin_addr.s_addr &
	   inet_addr (inet_ntoa
		      (localnet
		       [lnnum].subnetmask)))
	  == inet_addr (inet_ntoa (localnet[lnnum].subnetwork)))
	{
	  isonlocalnet = TCONN_TRUE;
	  break;
	}
    }
  if(TCONN_DEBUG) fprintf(stderr,"Connecting to %s\n"
		"isonlocalnet == %d\n"
		"proxyservset == %d\n"
		"proxyportset == %d\n"
		"localnetset  == %d\n",inet_ntoa(((struct sockaddr_in *)
		serv_addr)->sin_addr),isonlocalnet,proxyservset,
		proxyportset,localnetset);

  if (!isonlocalnet && proxyservset && proxyportset && localnetset)
    {				/* Throught proxy */
      struct sockaddr_in proxyaddr;
      char connectbuf[100];
      int connectbuflen;
      char authbuf[100];
      int authbuflen;
      char recvchar;
      char headerbuf[1000];
      int hbuflen;
	  void (*oldpipehandler)(int);
	  void (*oldtimehandler)(int);
      int flags;
      /* Ignore the signals for SIGPIPE and SIGALRM */
      oldpipehandler = signal (SIGPIPE, SIG_IGN);
      oldtimehandler = signal (SIGALRM, SIG_IGN);
      /* Make the function Blocking if not */
      /* This is being tried since pine does not work otherwise */
      flags = fcntl(sockfd, F_GETFL);
      fcntl(sockfd,F_SETFL,!O_NONBLOCK);

      proxyaddr.sin_family = AF_INET;
      proxyaddr.sin_port = htons (proxyport);
      proxyaddr.sin_addr.s_addr = inet_addr (proxyserv);
      bzero (&(proxyaddr.sin_zero), 8);

      /* Connect to the proxy:port */
      if (func
	  (sockfd, (struct sockaddr *) &proxyaddr,
	   sizeof (struct sockaddr)) != 0)
	{
	  signal(SIGPIPE,oldpipehandler);
	  signal(SIGALRM,oldtimehandler);
          fcntl(sockfd,F_SETFL,flags);
          fprintf(stderr,"Can't connect to proxy, See that your "
		"proxy settings are correct.\n");
	  errno = ECONNREFUSED;
	  return -1;
	}

      if (TCONN_DEBUG) fprintf (stderr, "Through proxy\n");
      connectbuflen = snprintf (connectbuf, 100, "CONNECT %s:%d HTTP/1.0\r\n",
				inet_ntoa (
					   ((struct sockaddr_in *)
					    serv_addr)->sin_addr),
				ntohs (
				       ((struct sockaddr_in *)
					serv_addr)->sin_port));

      if (send (sockfd, connectbuf, connectbuflen, 0) != connectbuflen)
	{
	  signal(SIGPIPE,oldpipehandler);
	  signal(SIGALRM,oldtimehandler);
          fcntl(sockfd,F_SETFL,flags);
	  errno = ECONNREFUSED;
	  return -1;
	}
      if (TCONN_DEBUG) fprintf (stderr, "%s", connectbuf);
      if (proxypassset && proxyuserset)
	{
           char authstring[60];
           strncpy(authstring,proxyuser,9);
	   strncat(authstring,":",2);
	   strncat(authstring,proxypass,19);
	   
	   if (TCONN_DEBUG) fprintf (stderr, " authstring is %s\n", authstring);
       	   ajayd_tconn_base64_encode(authstring,proxyauth);
	   authbuflen =
	    snprintf (authbuf, 100,
		      "Proxy-Authorization: Basic %s\r\n", proxyauth);

	  if (send (sockfd, authbuf, authbuflen, 0) != authbuflen)
	    {
	      signal(SIGPIPE,oldpipehandler);
	      signal(SIGALRM,oldtimehandler);
          fcntl(sockfd,F_SETFL,flags);
	      errno = ECONNREFUSED;
	      return -1;
	    }
	  if (TCONN_DEBUG) fprintf (stderr, "%s", authbuf);
	}

      if (send (sockfd, "\r\n", 2, 0) != 2)
	{
	  signal(SIGPIPE,oldpipehandler);
	  signal(SIGALRM,oldtimehandler);
          fcntl(sockfd,F_SETFL,flags);
	  errno = ECONNREFUSED;
	  return -1;
	}

      if (TCONN_DEBUG) fprintf (stderr, "\r\n");

      hbuflen = 0;

      while (1)
	{
	  if (recv (sockfd, &recvchar, 1, 0) != 1)
	    {
	      signal(SIGPIPE,oldpipehandler);
	      signal(SIGALRM,oldtimehandler);
          fcntl(sockfd,F_SETFL,flags);
	      errno = ETIMEDOUT;
	      return -1;
	    }

	  if (hbuflen == 999)	/* This means only junk is coming. */
	    {
	      signal(SIGPIPE,oldpipehandler);
	      signal(SIGALRM,oldtimehandler);
          fcntl(sockfd,F_SETFL,flags);
	      errno = ETIMEDOUT;
	      return -1;
	    }

	  /* Copy the header */
	  headerbuf[hbuflen] = recvchar;
	  if (hbuflen > 1
	      && (headerbuf[hbuflen] == '\n'
		  && headerbuf[hbuflen - 2] == '\n'))
	    break;
	  hbuflen++;
	}


	signal(SIGPIPE,oldpipehandler);
	signal(SIGALRM,oldtimehandler);
          fcntl(sockfd,F_SETFL,flags);

/* I don't know why but the header was not getting printed with only one '\0' */
      headerbuf[hbuflen + 1] = headerbuf[hbuflen + 2] = '\0';

      if (TCONN_DEBUG) fprintf (stderr, "\nHeaders are:\n%s", headerbuf);
      /* Hack this to fit to your proxy needs.
       * The squid here responds with "HTTP/1.? 200 Connection established"
       */
      if (atoi (headerbuf + 9) == 200)
	{
	  if (TCONN_DEBUG) fprintf (stderr, "Received 200 OK from proxy\n");
	  errno = 0;
	  return 0;
	}
      else if (atoi (headerbuf + 9) == 407)
	{
	  /* Shouldn't be so, but just for the heck of it. */
	  fprintf (stderr,
		   "Hey, Your proxy Authentication is not Working.\n");
	  errno = ECONNREFUSED;
	  return -1;
	}
      else
	{
	  if (TCONN_DEBUG) fprintf (stderr, "Connection Refused: Service Unavaliable\n");
	  errno = ECONNREFUSED;
	  return -1;
	}
    }

directconnect:
  {
    if (TCONN_DEBUG) fprintf (stderr, "Direct connection\n");
	/* Check if the dynamic linking is working */
	/* If it is not working func may be either NULL */
	/* or same as connect itself when it would go into */
	/* an infinite loop and cause segmentation fault */
	if(func != NULL  && func != connect )
    	return func (sockfd, serv_addr, addrlen);
	else
	{
		fprintf(stderr,"Dynamic Linking is not working\n If you are on BSD"
		"Did you specify the correct libc file during installation\n");
		errno = ENETUNREACH;
		return -1;
	}
  }
}

/* Just to make sure that the name does not interfere with any other function */

void ajayd_tconn_base64_encode (const char *s, char * p)
{
        char tbl[64] = {
                'A','B','C','D','E','F','G','H',
                'I','J','K','L','M','N','O','P',
                'Q','R','S','T','U','V','W','X',
                'Y','Z','a','b','c','d','e','f',
                'g','h','i','j','k','l','m','n',
                'o','p','q','r','s','t','u','v',
                'w','x','y','z','0','1','2','3',
                '4','5','6','7','8','9','+','/'
        };
        int i,length;
        
        length = strlen(s);

        for (i = 0; i < length; i += 3)
        {
                *p++ = tbl[s[0] >> 2];
                *p++ = tbl[((s[0] & 3) << 4) + (s[1] >> 4)];
                *p++ = tbl[((s[1] & 0xf) << 2) + (s[2] >> 6)];
                *p++ = tbl[s[2] & 0x3f];
                s += 3;
        }
        if (i == length + 1)
                *(p - 1) = '=';
        else if (i == length + 2)
                *(p - 1) = *(p - 2) = '=';
        *p = '\0';
}

